package org.jvnet.jaxbcommons.extendedcloneable.addon.generator;

import java.lang.reflect.Method;
import java.lang.reflect.Modifier;

import org.jvnet.jaxbcommons.cloneable.Cloneable;
import org.jvnet.jaxbcommons.lang.CloneAware;
import org.jvnet.jaxbcommons.util.TypeUtils;
import org.jvnet.jaxbcommons.visitor.ClassifyingVisitor;
import org.w3c.dom.Element;

import com.sun.codemodel.JAssignment;
import com.sun.codemodel.JAssignmentTarget;
import com.sun.codemodel.JBlock;
import com.sun.codemodel.JClass;
import com.sun.codemodel.JConditional;
import com.sun.codemodel.JExpr;
import com.sun.codemodel.JExpression;
import com.sun.codemodel.JForLoop;
import com.sun.codemodel.JOp;
import com.sun.codemodel.JStatement;
import com.sun.codemodel.JType;
import com.sun.codemodel.JVar;
import com.sun.tools.xjc.generator.ClassContext;
import com.sun.tools.xjc.grammar.FieldItem;
import com.sun.tools.xjc.grammar.PrimitiveItem;

public class EntryCopyVisitor extends ClassifyingVisitor {

  private final JExpression value;
  private final JAssignmentTarget valueCopy;
  private final JExpression clonedObjects;
  private final JType type;

  public EntryCopyVisitor(
      ClassContext classContext,
      JExpression value,
      JAssignmentTarget valueCopy,
      JExpression clonedObjects,
      JType type) {
    super(classContext);
    this.value = value;
    this.valueCopy = valueCopy;
    this.clonedObjects = clonedObjects;
    this.type = type;
  }

  public Object onPrimitiveSingleField(FieldItem fieldItem) {
    final PrimitiveItem primitiveItem = TypeUtils.getPrimitiveItem(fieldItem);
    return assignPrimitive(this.valueCopy, this.value, primitiveItem.getType());
  }

  private JStatement assignPrimitive(
      final JAssignmentTarget valueCopy,
      final JExpression value,
      final JType type) {
    if (type.isArray()) {
      final JBlock block = JBlock.dummyInstance.block();
      final JType elementType = type.elementType();
      block.assign(valueCopy, JExpr.newArray(elementType, value.ref("length")));
      final JForLoop _for = block._for();
      final JVar i = _for.init(codeModel.INT, "i", JExpr.lit(0));
      _for.test(JOp.lt(i, value.ref("length")));
      _for.update(JOp.incr(i));
      final JVar v = _for.body().decl(elementType, "v", value.component(i));
      assignPrimitive(valueCopy.component(i), v, elementType);
      return block;
    }
    else if (type instanceof JClass) {
      final JClass _class = (JClass) type;
      final String name = _class.fullName();
      try {
        final Class _theClass = Class.forName(name);
        try {
          final Method method = _theClass.getMethod("clone", null);
          if (method != null && Modifier.isPublic(method.getModifiers())) {
            return assignClone(valueCopy, value, type);
          }
          else {
            return assign(valueCopy, value);
          }
        }
        catch (NoSuchMethodException nsmex) {
          return assign(valueCopy, value);
        }
      }
      catch (ClassNotFoundException cnfex) {
        logger.warn("Class [" + name + "]of the primitive type must be known.");
        return assign(valueCopy, value);
      }
    }
    else {
      return assign(valueCopy, value);
    }
  }

  private JAssignment assign(final JAssignmentTarget valueCopy, final JExpression value) {
    return (JAssignment) valueCopy.assign(value);
  }

  private JAssignment assignClone(
      final JAssignmentTarget valueCopy,
      final JExpression value,
      JType type) {
    return (JAssignment) valueCopy.assign(JExpr.cast(type, value.invoke("clone")));
  }

  private JAssignment assignVisitClone(
      final JAssignmentTarget valueCopy,
      final JExpression value,
      JType type) {
    return (JAssignment) valueCopy.assign(JExpr.cast(type, value.invoke("visitClone").arg(
        clonedObjects)));
  }

  public Object onEnumSingleField(FieldItem fieldItem) {
    return assign(this.valueCopy, this.value);
  }

  public Object onDomSingleField(FieldItem fieldItem) {
    return assignDomCopy(this.valueCopy, this.value);
  }

  private JAssignment assignDomCopy(final JAssignmentTarget valueCopy, final JExpression value) {
    return (JAssignment) valueCopy.assign(JExpr.cast(codeModel.ref(Element.class), value.invoke(
        "cloneNode").arg(JExpr.TRUE)));
  }

  public Object onWildcardSingleField(FieldItem fieldItem) {
    return assignWildcardCopy(this.valueCopy, this.value, type);
  }

  private JStatement assignWildcardCopy(
      final JAssignmentTarget valueCopy,
      final JExpression value,
      final JType type) {
    final JConditional ifCloneAware = JBlock.dummyInstance._if(JOp._instanceof(value, codeModel
        .ref(CloneAware.class)));
    ifCloneAware._then().add(
        assignVisitClone(valueCopy, JExpr.cast(codeModel.ref(CloneAware.class), value), type));

    final JConditional ifCloneable = ifCloneAware._else()._if(
        JOp._instanceof(value, codeModel.ref(Cloneable.class)));

    ifCloneable._then().add(
        assignClone(valueCopy, JExpr.cast(codeModel.ref(Cloneable.class), value), type));
    ifCloneable._else().add(assign(valueCopy, value));
    return ifCloneAware;
  }

  public Object onComplexSingleField(FieldItem fieldItem) {
    final JType theType = TypeUtils.getCommonBaseTypeItem(classContext, fieldItem).getType();
    return assignVisitClone(this.valueCopy, this.value, theType);
  }

  public Object onHeteroSingleField(FieldItem fieldItem) {
    return assignWildcardCopy(valueCopy, value, type);
  }

  public Object onPrimitiveCollectionField(FieldItem fieldItem) {
    return onPrimitiveSingleField(fieldItem);
  }

  public Object onEnumCollectionField(FieldItem fieldItem) {
    return onEnumSingleField(fieldItem);
  }

  public Object onDomCollectionField(FieldItem fieldItem) {
    return onDomSingleField(fieldItem);
  }

  public Object onComplexCollectionField(FieldItem fieldItem) {
    return onComplexSingleField(fieldItem);
  }

  public Object onWildcardCollectionField(FieldItem fieldItem) {
    return onWildcardSingleField(fieldItem);
  }

  public Object onHeteroCollectionField(FieldItem fieldItem) {
    return onHeteroSingleField(fieldItem);
  }
}
