package org.jvnet.jaxbcommons.inheritance.addon;

import java.io.IOException;
import java.util.Collections;
import java.util.List;

import javax.xml.namespace.QName;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.jvnet.jaxbcommons.util.ClassUtils;
import org.jvnet.jaxbcommons.util.CodeModelUtils;
import org.jvnet.jaxbcommons.util.CustomizationUtils;
import org.xml.sax.ErrorHandler;

import com.sun.codemodel.JClass;
import com.sun.codemodel.JCodeModel;
import com.sun.tools.xjc.BadCommandLineException;
import com.sun.tools.xjc.CodeAugmenterEx;
import com.sun.tools.xjc.Options;
import com.sun.tools.xjc.generator.ClassContext;
import com.sun.tools.xjc.generator.GeneratorContext;
import com.sun.tools.xjc.grammar.AnnotatedGrammar;
import com.sun.tools.xjc.grammar.ClassItem;
import com.sun.tools.xjc.reader.xmlschema.bindinfo.BIXPluginCustomization;

public class AddOn implements CodeAugmenterEx {

	private static final QName EXTENDS_ELEMENT_NAME = new QName(
			"http://jaxbcommons.dev.java.net/inheritance", "extends");

	private static final QName IMPLEMENTS_ELEMENT_NAME = new QName(
			"http://jaxbcommons.dev.java.net/inheritance", "implements");

	/**
	 * Logger.
	 */
	protected Log logger = LogFactory.getLog(AddOn.class);

	/**
	 * Does nothing.
	 * 
	 * @param opt
	 *            ignored.
	 * @param args
	 *            ignored.
	 * @param i
	 *            ignored.
	 * @return Alway returns <code>0</code>.
	 * @throws com.sun.tools.xjc.BadCommandLineException
	 *             Never thrown.
	 * @throws java.io.IOException
	 *             Never thrown.
	 */
	public int parseArgument(final Options opt, final String[] args, final int i)
			throws BadCommandLineException, IOException {
		return 0; // no option recognized
	}

	public String getOptionName() {
		return "Xinheritance";
	}

	public String getUsage() {
		return "  -Xinheritance                :  ";
	}

	/**
	 * Augments the code.
	 * 
	 * @param grammar
	 *            annotated grammar.
	 * @param generatorContext
	 *            generator context.
	 * @param options
	 *            options.
	 * @param errorHandler
	 *            error handler.
	 * @return If the add-on executes successfully, return true. If it detects
	 *         some errors but those are reported and recovered gracefully,
	 *         return false.
	 */
	public boolean run(final AnnotatedGrammar grammar,
			final GeneratorContext generatorContext, final Options options,
			final ErrorHandler errorHandler) {

		logger.debug("Starting code augmentation.");

		// Process class items
		final ClassItem[] classItems = grammar.getClasses();
		for (int index = 0; index < classItems.length; index++) {
			final ClassItem classItem = classItems[index];
			final ClassContext classContext = generatorContext
					.getClassContext(classItem);
			logger.debug("Processing [" + classItem.name + "].");
			process(classContext);
		}
		logger.debug("Finished code augmentation.");
		return true;
	}

	protected void process(ClassContext classContext) {

		final JCodeModel codeModel = CodeModelUtils.getCodeModel(classContext);
		final String _implements = getImplements(classContext);

		if (_implements != null) {
			try {
				final JClass _implemented = codeModel.ref(_implements);
				logger.info("Interface [" + classContext.ref.fullName()
						+ "] will extend [" + _implemented.fullName() + "].");
				ClassUtils._implements(classContext.ref, _implemented);

				logger
						.info("Class [" + classContext.implClass.fullName()
								+ "] will implement ["
								+ _implemented.fullName() + "].");
				ClassUtils._implements(classContext.implClass, _implemented);

			} catch (ClassNotFoundException cnfex) {
				logger.error("Could not resolve the interface [" + _implements
						+ "].", cnfex);
			}
		}

		final String _extends = getExtends(classContext);
		if (_extends != null) {
			try {
				final JClass _extended = codeModel.ref(_extends);
				if (classContext.implClass._extends() != null
						&& !codeModel.ref(Object.class).equals(
								classContext.implClass._extends())) {
					logger.warn("Class [" + classContext.implClass.fullName()
							+ "] already extends ["
							+ classContext.implClass._extends().fullName()
							+ "].");
				} else {
					logger.info("Class [" + classContext.implClass.fullName()
							+ "] will extend [" + _extended.fullName() + "].");
					classContext.implClass._extends(_extended);
				}
			} catch (ClassNotFoundException cnfex) {
				logger.error("Could not resolve the class[" + _extends + "].",
						cnfex);
			}

		}
	}

	private String getExtends(ClassContext classContext) {
		final BIXPluginCustomization customization = CustomizationUtils
				.getCustomization(classContext.target, EXTENDS_ELEMENT_NAME);
		if (customization != null) {
			return customization.element.getAttribute("class");
		} else {
			return null;
		}
	}

	private String getImplements(ClassContext classContext) {
		final BIXPluginCustomization customization = CustomizationUtils
				.getCustomization(classContext.target, IMPLEMENTS_ELEMENT_NAME);
		if (customization != null) {
			return customization.element.getAttribute("interface");
		} else {
			return null;
		}
	}

	public List getCustomizationURIs() {
		return Collections
				.singletonList("http://jaxbcommons.dev.java.net/inheritance");
	}
}
